import os
import sys
import glob
import yaml
import logging
import unicodedata

import hydra
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from tqdm.auto import tqdm


@hydra.main(version_base="1.2", config_path="translation_config", config_name="retranslate")
def main(cfg):
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
        level=logging.INFO,
    )

    logging.info("Loading re-translation model")
    logging.debug("  Loading tokenizer")
    tokenizer = AutoTokenizer.from_pretrained(cfg.model.hf_model_name, src_lang="eng_Latn")
    if cfg.output.list_language_codes:
        for lang_code in sorted(list(tokenizer.lang_code_to_id.keys())):
            print(lang_code)
        return
    logging.debug("  Loading model")
    model = AutoModelForSeq2SeqLM.from_pretrained(cfg.model.hf_model_name)
    if cfg.model.cuda:
        logging.debug("  Copying model to GPU")
        model.cuda()

    if cfg.generation.word_exclusion_list is not None:
        if not os.path.isfile(cfg.generation.word_exclusion_list) or \
            not os.path.splitext(cfg.generation.word_exclusion_list)[1] in {'.yaml', '.yml'} \
        :
            logging.error("The word exclusion list should be a YAML file")
            raise FileNotFoundError("The word exclusion list should be a YAML file")
        logging.info("Tokenizing word exclusion list")
        with open(cfg.generation.word_exclusion_list, 'r') as f:
            bad_words = yaml.safe_load(f)
        bad_words = bad_words.get(cfg.output.target_language_code, [])
        bad_word_tokens = [tokenizer.encode(w, add_special_tokens=False) for w in tqdm(bad_words)]
    else:
        bad_word_tokens = []

    if getattr(cfg.generation, "alphabet", None) is not None:
        logging.info("Preparing alphabet constraints")
        if not cfg.generation.alphabet in cfg.generation.known_alphabets:
            logging.warning(f"Provided alphabet name '{cfg.generation.alphabet}' not in set of known alphabets. "
                             "Will attempt to generate anyway but you may get only symbols in your output and the "
                             "generation may take an extremely long time, even with GPU acceleration switched on. "
                             "The list of known alphabets is: "
                            f"{', '.join(sorted(list(cfg.generation.known_alphabets.keys())))}.")
        alphabet_name = cfg.generation.known_alphabets.get(cfg.generation.alphabet, cfg.generation.alphabet).lower()
        base_inclusions = set(cfg.generation.base_alphabet_inclusions)
        allowed_token_ids = tokenizer.convert_tokens_to_ids(filter(
            lambda t: all([
                c in base_inclusions or alphabet_name in unicodedata.name(c, "").lower() for c in t
            ]),
            tokenizer.vocab.keys()
        )) + [tokenizer.eos_token_id]
        token_constraints_fn = lambda _1, _2: allowed_token_ids
    else:
        token_constraints_fn = None

    logging.info("Re-translating")
    output_ext = getattr(cfg.output, "output_extension", cfg.output.target_language_code)
    in_fns = glob.glob(cfg.input_glob)
    for in_fn in (files_pbar := tqdm(in_fns)):
        files_pbar.set_description(os.path.basename(in_fn))
        with open(in_fn, 'r') as f_in:
            translated = [l.strip() for l in f_in.readlines()]
        out_fn = os.path.join(cfg.output.output_dir, os.path.splitext(os.path.basename(in_fn))[0] + '.' + output_ext)
        with open(out_fn, 'w') as f_out:
            for line in tqdm(translated, leave=False, desc="Translating"):
                if any('TIBETAN' in unicodedata.name(c) for c in line):
                    f_out.write(line + '\n')
                elif len(line) == 0:
                    f_out.write('\n')
                else:
                    translated_tokens = model.generate(
                        **tokenizer(line, return_tensors="pt").to(model.device),
                        forced_bos_token_id=tokenizer.lang_code_to_id[cfg.output.target_language_code],
                        prefix_allowed_tokens_fn=token_constraints_fn,
                        bad_words_ids=None if len(bad_word_tokens) == 0 else bad_word_tokens,   # Needs to be None instead of empty, otherwise HF throws ValueError
                        max_length=cfg.model.max_length,
                        num_beams=getattr(cfg.generation, "num_beams", 10),
                        **dict(getattr(cfg.generation, "generator_kwargs", {}))
                    )[0]
                    f_out.write(tokenizer.decode(translated_tokens, skip_special_tokens=True) + '\n')
                f_out.flush()


if __name__ == "__main__":
    main()      # pylint: disable=no-value-for-parameter



import os
import sys
import copy
import logging
import glob

import hydra
from hydra.utils import instantiate
from omegaconf import OmegaConf
from tqdm.auto import tqdm

from cai_garland.utils.translator import Translator


def interactive(translator, _mode_cfg, _generation_cfg, target_language_code):
    print("Interactive Tibetan translation...")
    while True:
        print("===")
        bo_text = input("Tibetan (or type exit): ")
        if bo_text == "exit":
            break
        print(translator.translate(bo_text, target_language_code=target_language_code))


def batch(translator, mode_cfg, generation_cfg, target_language_code):
    if mode_cfg.input_glob is None:
        raise ValueError("Specify an input file (or glob) in the mode.input_glob setting. You can do this from the "
                         "command line.")
    os.makedirs(mode_cfg.output_dir, exist_ok=True)
    in_fns = glob.glob(mode_cfg.input_glob)
    original_cfg = copy.deepcopy(generation_cfg)
    for in_fn in (files_pbar := tqdm(in_fns)):
        files_pbar.set_description(os.path.basename(in_fn))
        with open(in_fn, encoding=mode_cfg.input_encoding) as in_f:
            bo_text = in_f.read()

        generation_cfg = copy.deepcopy(original_cfg)
        in_cfg_fn = os.path.join(os.path.dirname(in_fn), os.path.splitext(os.path.basename(in_fn))[0] + '.config.yaml')
        if os.path.isfile(in_cfg_fn):
            in_cfg = OmegaConf.load(in_cfg_fn)
            generation_cfg = OmegaConf.merge(generation_cfg, in_cfg)

        out_fn = os.path.join(
            mode_cfg.output_dir, os.path.splitext(os.path.basename(in_fn))[0] + '.' + mode_cfg.output_extension)
        translation_kwargs = {}
        translation_kwargs["retrospective_decoding"] = generation_cfg.generation.get("retrospective_decoding", False)
        translation_kwargs["retrospective_registers"] = \
            translator.model.encoder.config.model_type == "siamese-encoder" \
            and generation_cfg.generation.get("use_registers_for_retrospective_decoding", True)
        translation_kwargs["retrospection_window"] = generation_cfg.generation.get("retrospection_window", None)
        translation_kwargs["contextual_decoding"] = hasattr(generation_cfg.generation, "pooled_context")
        if translation_kwargs["contextual_decoding"]:
            translation_kwargs["context_window_words"] = generation_cfg.generation.pooled_context.context_window.words
            translation_kwargs["context_window_characters"] = \
                generation_cfg.generation.pooled_context.context_window.characters

        with open(out_fn, mode='w') as out_f:
            translator.hard_segmenter = instantiate(generation_cfg.segmentation.hard_segmentation)
            translator.preprocessors = [
                instantiate(preproc_func)
                for preproc_func in generation_cfg.processing.preprocessing
            ]
            translator.soft_segmenter = instantiate(
                generation_cfg.segmentation.soft_segmentation, translator=translator)
            translator.soft_segment_combiner_config = getattr(
                generation_cfg.segmentation, "soft_segment_combiner", None)
            translator.soft_segment_preprocessors = [
                instantiate(preproc_func)
                for preproc_func in generation_cfg.processing.get("soft_segment_preprocessing", [])
            ]
            translator.postprocessors = [
                instantiate(preproc_func)
                for preproc_func in generation_cfg.processing.postprocessing
            ]

            for src_segment, tgt_segment in translator.batch_translate(
                bo_text,
                tqdm=tqdm,
                hard_segmenter_kwargs=dict(generation_cfg.segmentation.hard_segmenter_kwargs),
                soft_segmenter_kwargs=dict(generation_cfg.segmentation.soft_segmenter_kwargs),
                throw_translation_errors=not mode_cfg.skip_long_inputs,
                target_language_code=target_language_code,
                generator_kwargs=dict(generation_cfg.generation.get("generator_kwargs", {})),
                **translation_kwargs
            ):
                if mode_cfg.output_parallel_translation:
                    out_f.write(src_segment + '\n')
                out_f.write('\n')
                out_f.write(tgt_segment + '\n')
                out_f.write('\n')
                out_f.flush()


@hydra.main(version_base="1.2", config_path="translation_config", config_name="translate")
def main(cfg):
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
        level=logging.INFO,
    )

    translator = Translator(os.path.join(cfg.model.model_ckpt, cfg.model.model_size))
    translator.num_beams = cfg.generation.generation.num_beams
    if hasattr(cfg.model, "decoding_length"):
        translator.decoding_length = cfg.model.decoding_length
    if hasattr(cfg.generation.generation, "pooled_context"):
        translator.prepare_context_encoder(cfg.generation.generation.pooled_context.context_encoder.hf_model_name)
    target_language_code = getattr(cfg, "target_language_code", None)

    if cfg.cuda:
        translator.cuda()

    if target_language_code is not None and hasattr(cfg, "word_exclusion"):
        translator.bad_words = getattr(cfg.word_exclusion, target_language_code, [])

    instantiate(cfg.mode.process_func, translator, cfg.mode, cfg.generation, target_language_code)


if __name__ == "__main__":
    main()      # pylint: disable=no-value-for-parameter



import os
import sys
import glob
import hydra
import pickle
import logging

from tqdm.auto import tqdm
from hydra.utils import instantiate
from cai_manas.part_of_speech.pos_tagger import PartOfSpeechTagger


def interactive(tagger, _mode_cfg):
    print("Interactive Tibetan part-of-speech tagging...")
    while True:
        print("===")
        bo_text = input("Tibetan (or type exit): ")
        if bo_text == "exit":
            break
        res = tagger.tag(bo_text)
        for word, tag in zip(res["words"], res["tags"]):
            print(word, tag)


def batch(tagger, mode_cfg):
    if mode_cfg.input_glob is None:
        raise ValueError("Specify an input file (or glob) in the mode.input_glob setting. You can do this from the "
                         "command line.")
    to_pickle = mode_cfg.output_pickle_file
    os.makedirs(mode_cfg.output_dir, exist_ok=True)
    in_fns = sorted(glob.glob(mode_cfg.input_glob))
    logging.info("Processing input glob")
    for in_fn in (files_pbar := tqdm(in_fns)):
        files_pbar.set_description(os.path.basename(in_fn))
        with open(in_fn, encoding=mode_cfg.input_encoding) as in_f:
            bo_text = in_f.read()

        out_fn = os.path.join(
            mode_cfg.output_dir, os.path.splitext(os.path.basename(in_fn))[0] + '.' + mode_cfg.output_extension)

        if to_pickle:
            for_pickle = []
        with open(out_fn, mode='w') as out_f:
            tagger.segmenter = instantiate(mode_cfg.segmenter.segmenter, translator=tagger).to(tagger.model.device)
            tagger.preprocessors = [instantiate(preproc_func) for preproc_func in mode_cfg.get("preprocessing", [])]

            for segments, tags in tagger.batch_tag(
                bo_text,
                tqdm=tqdm,
                segmenter_kwargs=dict(mode_cfg.segmenter.kwargs),
                throw_encoder_errors=not mode_cfg.skip_long_inputs
            ):
                out_f.write(' '.join([f"{segment}[{tag}]" for segment, tag in zip(segments, tags)]))
                out_f.write('\n')
                out_f.write('\n')
                out_f.flush()
                if to_pickle:
                    for_pickle.append((segments, tags))
        if to_pickle:
            with open(out_fn + '.pkl', 'wb') as out_f:
                pickle.dump(for_pickle, out_f)


@hydra.main(version_base="1.2", config_path="token_classification_config", config_name="part_of_speech")
def main(cfg):
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
        level=logging.INFO,
    )

    logging.info("Loading model")
    tagger = PartOfSpeechTagger(cfg.model.model_ckpt)
    if cfg.cuda:
        logging.info("Moving model to GPU")
        tagger.cuda()

    instantiate(cfg.mode.process_func, tagger, cfg.mode)


if __name__ == "__main__":
    main()      # pylint: disable=no-value-for-parameter
